@@ -2,7 +2,7 @@
  * The transitional conversion from the output of the Python-based parser to
  * that required by the rest of the C-based code generator.
  *
- * Copyright (c) 2022 Riverbank Computing Limited <info@riverbankcomputing.com>
+ * Copyright (c) 2023 Riverbank Computing Limited <info@riverbankcomputing.com>
  *
  * This file is part of SIP.
  *
@@ -100,7 +100,8 @@ static objectCache *cache_qual = NULL;
 static objectCache *cache_virtualerrorhandler = NULL;
 static objectCache *cache_wrappedenum = NULL;
 static objectCache *cache_wrappedtypedef = NULL;
-static strCache *cache_typehint = NULL;
+static strCache *cache_typehint_in = NULL;
+static strCache *cache_typehint_out = NULL;
 
 
 /*
@@ -121,7 +122,8 @@ static void clear_caches(void)
     clear_object_cache(&cache_virtualerrorhandler);
     clear_object_cache(&cache_wrappedenum);
     clear_object_cache(&cache_wrappedtypedef);
-    clear_str_cache(&cache_typehint);
+    clear_str_cache(&cache_typehint_in);
+    clear_str_cache(&cache_typehint_out);
 }
 
 
@@ -234,9 +236,10 @@ static void *search_str_cache(const strCache *cache, char *py_str)
 
 
 /* Forward declarations of convertors. */
-static argDef *argument(sipSpec *pt, PyObject *obj, const char *encoding);
-static argDef *argument_attr(sipSpec *pt, PyObject *obj, const char *name,
-        const char *encoding);
+static void argument(sipSpec *pt, PyObject *obj, const char *encoding,
+        argDef *value);
+static void argument_attr(sipSpec *pt, PyObject *obj, const char *name,
+        const char *encoding, argDef *value);
 static argList *argument_list_attr(sipSpec *pt, PyObject *obj,
         const char *name, const char *encoding);
 static int bool_attr(PyObject *obj, const char *name);
@@ -298,8 +301,6 @@ static memberDef *member_list_attr(sipSpec *pt, PyObject *obj,
 static moduleDef *module(sipSpec *pt, PyObject *obj, const char *encoding);
 static moduleDef *module_attr(sipSpec *pt, PyObject *obj, const char *name,
         const char *encoding);
-static moduleDef *module_list_attr(sipSpec *pt, PyObject *obj,
-        const char *name, const char *encoding);
 static moduleListDef *modulelist_attr(sipSpec *pt, PyObject *obj,
         const char *name, const char *encoding);
 static mroDef *mro_list_attr(sipSpec *pt, PyObject *obj, const char *name,
@@ -319,9 +320,9 @@ static scopedNameDef *scopedname(PyObject *obj, const char *encoding);
 static scopedNameDef *scopedname_attr(PyObject *obj, const char *name,
         const char *encoding);
 static signatureDef *signature(sipSpec *pt, PyObject *obj,
-        const char *encoding);
+        const char *encoding, signatureDef *value);
 static signatureDef *signature_attr(sipSpec *pt, PyObject *obj,
-        const char *name, const char *encoding);
+        const char *name, const char *encoding, signatureDef *value);
 static sourceLocation sourcelocation(PyObject *obj, const char *encoding);
 static sourceLocation sourcelocation_attr(PyObject *obj, const char *name,
         const char *encoding);
@@ -337,7 +338,7 @@ static throwArgs *throw_arguments(sipSpec *pt, PyObject *obj,
 static throwArgs *throw_arguments_attr(sipSpec *pt, PyObject *obj,
         const char *name, const char *encoding);
 static typeHintDef *typehint_attr(PyObject *obj, const char *name,
-        const char *encoding);
+        const char *encoding, strCache **cache);
 static void typehints_attr(PyObject *obj, const char *name,
         const char *encoding, typeHintDef **th_in, typeHintDef **th_out,
         const char **th_value);
@@ -398,8 +399,7 @@ sipSpec *py2c(PyObject *spec, const char *encoding)
 
     pt = sipMalloc(sizeof (sipSpec));
 
-    pt->modules = module_list_attr(pt, spec, "modules", encoding);
-    pt->module = pt->modules;
+    pt->module = module_attr(pt, spec, "module", encoding);
     pt->namecache = cachedname_list_attr(spec, "name_cache", encoding);
     pt->ifacefiles = ifacefile_list_attr(pt, spec, "iface_files", encoding);
     pt->classes = class_list_attr(pt, spec, "classes", encoding);
@@ -417,6 +417,7 @@ sipSpec *py2c(PyObject *spec, const char *encoding)
     pt->exptypehintcode = codeblock_list_attr(spec, "exported_type_hint_code",
             encoding);
     pt->genc = bool_attr(spec, "c_bindings");
+    pt->is_composite = bool_attr(spec, "is_composite");
     pt->plugins = str_list_attr(spec, "plugins", encoding);
     pt->nrvirthandlers = int_attr(spec, "nr_virtual_handlers");
     pt->qobject_cd = class_attr(pt, spec, "pyqt_qobject", encoding);
@@ -428,9 +429,9 @@ sipSpec *py2c(PyObject *spec, const char *encoding)
 /*
  * Convert an optional Argument object.
  */
-static argDef *argument(sipSpec *pt, PyObject *obj, const char *encoding)
+static void argument(sipSpec *pt, PyObject *obj, const char *encoding,
+        argDef *value)
 {
-    argDef *value = sipMalloc(sizeof (argDef));
     PyObject *derefs_obj, *definition;
     int key, xfer, array;
     Py_ssize_t i;
@@ -438,7 +439,7 @@ static argDef *argument(sipSpec *pt, PyObject *obj, const char *encoding)
     if (obj == Py_None)
     {
         value->atype = no_type;
-        return value;
+        return;
     }
 
     value->atype = (argType)enum_attr(obj, "type");
@@ -523,7 +524,7 @@ static argDef *argument(sipSpec *pt, PyObject *obj, const char *encoding)
     switch (value->atype)
     {
     case function_type:
-        value->u.sa = signature(pt, definition, encoding);
+        value->u.sa = signature(pt, definition, encoding, NULL);
         break;
 
     case template_type:
@@ -560,27 +561,22 @@ static argDef *argument(sipSpec *pt, PyObject *obj, const char *encoding)
     }
 
     Py_DECREF(definition);
-
-    return value;
 }
 
 
 /*
  * Convert an optional Argument attribute.
  */
-static argDef *argument_attr(sipSpec *pt, PyObject *obj, const char *name,
-        const char *encoding)
+static void argument_attr(sipSpec *pt, PyObject *obj, const char *name,
+        const char *encoding, argDef *value)
 {
     PyObject *attr = PyObject_GetAttrString(obj, name);
-    argDef *value;
 
     assert(attr != NULL);
 
-    value = argument(pt, attr, encoding);
+    argument(pt, attr, encoding, value);
 
     Py_DECREF(attr);
-
-    return value;
 }
 
 
@@ -603,7 +599,7 @@ static argList *argument_list_attr(sipSpec *pt, PyObject *obj,
     {
         argList *item = sipMalloc(sizeof (argList));
 
-        item->arg = *argument(pt, PyList_GetItem(attr, i), encoding);
+        argument(pt, PyList_GetItem(attr, i), encoding, &item->arg);
 
         *tail = item;
         tail = &item->next;
@@ -863,6 +859,7 @@ static classDef *class(sipSpec *pt, PyObject *obj, const char *encoding)
             encoding);
     typehints_attr(obj, "type_hints", encoding, &value->typehint_in,
             &value->typehint_out, &value->typehint_value);
+    value->len_cpp_name = str_attr(obj, "len_cpp_name", encoding);
 
     return value;
 }
@@ -1046,7 +1043,7 @@ static ctorDef *constructor(sipSpec *pt, PyObject *obj, const char *encoding)
     py_sig_obj = PyObject_GetAttrString(obj, "py_signature");
     assert(py_sig_obj != NULL);
 
-    value->pysig = *signature(pt, py_sig_obj, encoding);
+    signature(pt, py_sig_obj, encoding, &value->pysig);
 
     cpp_sig_obj = PyObject_GetAttrString(obj, "cpp_signature");
     assert(cpp_sig_obj != NULL);
@@ -1056,7 +1053,7 @@ static ctorDef *constructor(sipSpec *pt, PyObject *obj, const char *encoding)
         if (cpp_sig_obj == py_sig_obj)
             value->cppsig = &value->pysig;
         else
-            value->cppsig = signature(pt, cpp_sig_obj, encoding);
+            value->cppsig = signature(pt, cpp_sig_obj, encoding, NULL);
     }
 
     Py_DECREF(py_sig_obj);
@@ -1272,7 +1269,7 @@ static fcallDef *functioncall(sipSpec *pt, PyObject *obj, const char *encoding)
     PyObject *args_obj;
     Py_ssize_t i;
 
-    value->type = *argument_attr(pt, obj, "result", encoding);
+    argument_attr(pt, obj, "result", encoding, &value->type);
 
     args_obj = PyObject_GetAttrString(obj, "args");
     assert(args_obj != NULL);
@@ -1482,7 +1479,7 @@ static mappedTypeDef *mappedtype(sipSpec *pt, PyObject *obj,
     if (bool_attr(obj, "needs_user_state"))
         setNeedsUserState(value);
 
-    value->type = *argument_attr(pt, obj, "type", encoding);
+    argument_attr(pt, obj, "type", encoding, &value->type);
     value->pyname = cachedname_attr(obj, "py_name", encoding);
     value->cname = cachedname_attr(obj, "cpp_name", encoding);
     typehints_attr(obj, "type_hints", encoding, &value->typehint_in,
@@ -1659,12 +1656,6 @@ static moduleDef *module(sipSpec *pt, PyObject *obj, const char *encoding)
     if (bool_attr(obj, "has_delayed_dtors"))
         setHasDelayedDtors(value);
 
-    if (bool_attr(obj, "is_composite"))
-    {
-        setIsComposite(value);
-        value->modflags &= ~MOD_SUPER_INIT_MASK;
-    }
-
     if (bool_attr(obj, "use_arg_names"))
         setUseArgNames(value);
 
@@ -1704,7 +1695,6 @@ static moduleDef *module(sipSpec *pt, PyObject *obj, const char *encoding)
     value->next_key = int_attr(obj, "next_key");
     value->license = license_attr(obj, "license", encoding);
     value->proxies = class_list_attr(pt, obj, "proxies", encoding);
-    value->container = module_attr(pt, obj, "composite", encoding);
     value->used = ifacefilelist_attr(pt, obj, "used", encoding);
     value->imports = modulelist_attr(pt, obj, "imports", encoding);
     value->allimports = modulelist_attr(pt, obj, "all_imports", encoding);
@@ -1719,8 +1709,8 @@ static moduleDef *module(sipSpec *pt, PyObject *obj, const char *encoding)
         value->needed_types = sipCalloc(nr_needed_types, sizeof (argDef));
 
         for (i = 0; i < nr_needed_types; ++i)
-            value->needed_types[i] = *argument(pt,
-                    PyList_GetItem(needed_types_obj, i), encoding);
+            argument(pt, PyList_GetItem(needed_types_obj, i), encoding,
+                    &value->needed_types[i]);
 
         value->nr_needed_types = nr_needed_types;
     }
@@ -1750,34 +1740,6 @@ static moduleDef *module_attr(sipSpec *pt, PyObject *obj, const char *name,
 }
 
 
-/*
- * Convert a Module list attribute.
- */
-static moduleDef *module_list_attr(sipSpec *pt, PyObject *obj,
-        const char *name, const char *encoding)
-{
-    PyObject *attr = PyObject_GetAttrString(obj, name);
-    moduleDef *head, **tail;
-    Py_ssize_t i;
-
-    assert(attr != NULL);
-
-    head = NULL;
-    tail = &head;
-
-    for (i = 0; i < PyList_Size(attr); ++i)
-    {
-        moduleDef *item = module(pt, PyList_GetItem(attr, i), encoding);
-        *tail = item;
-        tail = &item->next;
-    }
-
-    Py_DECREF(attr);
-
-    return head;
-}
-
-
 /*
  * Convert a Module list attribute to a moduleListDef.
  */
@@ -1938,7 +1900,7 @@ static overDef *over(sipSpec *pt, PyObject *obj, const char *encoding)
     py_sig_obj = PyObject_GetAttrString(obj, "py_signature");
     assert(py_sig_obj != NULL);
 
-    value->pysig = *signature(pt, py_sig_obj, encoding);
+    signature(pt, py_sig_obj, encoding, &value->pysig);
 
     cpp_sig_obj = PyObject_GetAttrString(obj, "cpp_signature");
     assert(cpp_sig_obj != NULL);
@@ -1946,7 +1908,7 @@ static overDef *over(sipSpec *pt, PyObject *obj, const char *encoding)
     if (cpp_sig_obj == py_sig_obj)
         value->cppsig = &value->pysig;
     else
-        value->cppsig = signature(pt, cpp_sig_obj, encoding);
+        value->cppsig = signature(pt, cpp_sig_obj, encoding, NULL);
 
     Py_DECREF(py_sig_obj);
     Py_DECREF(cpp_sig_obj);
@@ -2170,25 +2132,25 @@ static scopedNameDef *scopedname_attr(PyObject *obj, const char *name,
  * Convert an optional Signature object.
  */
 static signatureDef *signature(sipSpec *pt, PyObject *obj,
-        const char *encoding)
+        const char *encoding, signatureDef *value)
 {
-    signatureDef *value;
     PyObject *args_obj;
     Py_ssize_t i;
 
     if (obj == Py_None)
         return NULL;
 
-    value = sipMalloc(sizeof (signatureDef));
+    if (value == NULL)
+        value = sipMalloc(sizeof (signatureDef));
 
-    value->result = *argument_attr(pt, obj, "result", encoding);
+    argument_attr(pt, obj, "result", encoding, &value->result);
 
     args_obj = PyObject_GetAttrString(obj, "args");
     assert(args_obj != NULL);
     assert(PyList_Check(args_obj));
 
     for (i = 0; i < PyList_Size(args_obj) && i < MAX_NR_ARGS; ++i)
-        value->args[i] = *argument(pt, PyList_GetItem(args_obj, i), encoding);
+        argument(pt, PyList_GetItem(args_obj, i), encoding, &value->args[i]);
 
     value->nrArgs = i;
 
@@ -2202,14 +2164,13 @@ static signatureDef *signature(sipSpec *pt, PyObject *obj,
  * Convert an optional Signature attribute.
  */
 static signatureDef *signature_attr(sipSpec *pt, PyObject *obj,
-        const char *name, const char *encoding)
+        const char *name, const char *encoding, signatureDef *value)
 {
     PyObject *attr = PyObject_GetAttrString(obj, name);
-    signatureDef *value;
 
     assert(attr != NULL);
 
-    value = signature(pt, attr, encoding);
+    value = signature(pt, attr, encoding, value);
 
     Py_DECREF(attr);
 
@@ -2337,7 +2298,7 @@ static templateDef *template(sipSpec *pt, PyObject *obj, const char *encoding)
     value = sipMalloc(sizeof (templateDef));
 
     value->fqname = scopedname_attr(obj, "cpp_name", encoding);
-    value->types = *signature_attr(pt, obj, "types", encoding);
+    signature_attr(pt, obj, "types", encoding, &value->types);
 
     return value;
 }
@@ -2422,7 +2383,7 @@ static throwArgs *throw_arguments_attr(sipSpec *pt, PyObject *obj,
  * Convert an optional str attribute as a typeHintDef.
  */
 static typeHintDef *typehint_attr(PyObject *obj, const char *name,
-        const char *encoding)
+        const char *encoding, strCache **cache)
 {
     PyObject *attr = PyObject_GetAttrString(obj, name);
     typeHintDef *value;
@@ -2432,11 +2393,11 @@ static typeHintDef *typehint_attr(PyObject *obj, const char *name,
 
     if ((raw_hint = str(attr, encoding)) != NULL)
     {
-        if ((value = search_str_cache(cache_typehint, raw_hint)) == NULL)
+        if ((value = search_str_cache(*cache, raw_hint)) == NULL)
         {
             value = sipMalloc(sizeof (typeHintDef));
 
-            cache_str(&cache_typehint, raw_hint, value);
+            cache_str(cache, raw_hint, value);
 
             value->status = needs_parsing;
             value->raw_hint = raw_hint;
@@ -2466,8 +2427,9 @@ static void typehints_attr(PyObject *obj, const char *name,
 
     if (attr != Py_None)
     {
-        *th_in = typehint_attr(attr, "hint_in", encoding);
-        *th_out = typehint_attr(attr, "hint_out", encoding);
+        *th_in = typehint_attr(attr, "hint_in", encoding, &cache_typehint_in);
+        *th_out = typehint_attr(attr, "hint_out", encoding,
+                &cache_typehint_out);
         *th_value = str_attr(attr, "default_value", encoding);
     }
 
@@ -2678,8 +2640,8 @@ static virtHandlerDef *virtualhandler(sipSpec *pt, PyObject *obj,
 
     value = sipMalloc(sizeof (virtHandlerDef));
 
-    value->cppsig = signature_attr(pt, obj, "cpp_signature", encoding);
-    value->pysig = signature_attr(pt, obj, "py_signature", encoding);
+    value->cppsig = signature_attr(pt, obj, "cpp_signature", encoding, NULL);
+    value->pysig = signature_attr(pt, obj, "py_signature", encoding, NULL);
     value->virtcode = codeblock_list_attr(obj, "virtual_catcher_code",
             encoding);
     value->veh = virtualerrorhandler_attr(pt, obj, "virtual_error_handler",
@@ -3031,7 +2993,7 @@ static typedefDef *wrappedtypedef(sipSpec *pt, PyObject *obj,
     value->fqname = scopedname_attr(obj, "fq_cpp_name", encoding);
     value->ecd = class_attr(pt, obj, "scope", encoding);
     value->module = module_attr(pt, obj, "module", encoding);
-    value->type = *argument_attr(pt, obj, "type", encoding);
+    argument_attr(pt, obj, "type", encoding, &value->type);
 
     return value;
 }
@@ -3107,7 +3069,7 @@ static varDef *wrappedvariable(sipSpec *pt, PyObject *obj,
     if (bool_attr(obj, "needs_handler"))
         setNeedsHandler(value);
 
-    value->type = *argument_attr(pt, obj, "type", encoding);
+    argument_attr(pt, obj, "type", encoding, &value->type);
     value->accessfunc = codeblock_list_attr(obj, "access_code", encoding);
     value->getcode = codeblock_list_attr(obj, "get_code", encoding);
     value->setcode = codeblock_list_attr(obj, "set_code", encoding);
